#!/usr/bin/python
# -*- coding: utf-8 -*-

# This file is part of vdr-webvideo-plugin.
#
# Copyright 2009 Antti Ajanki <antti.ajanki@iki.fi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest
import subprocess
import os
import signal
import time
import imp
import urlparse
import httplib

webvi = imp.load_source('webvi', '../webvi/webvi')

class TestServiceModules(unittest.TestCase):
    def setUp(self):
        # Start the server process.
        sp = subprocess.Popen(['/usr/bin/python', '../webvid', '-r../webvidaemon/services'])
        self.serverpid = sp.pid
        # Wait a bit for the server to initialize itself.
        time.sleep(0.5)
        if sp.poll() is not None:
            raise RuntimeError('The server process failed to start')

        self.client = webvi.WVClient('127.0.0.1', 2357, None, [], [])
    
    def tearDown(self):
        self.client.close()
        os.kill(self.serverpid, signal.SIGTERM)
        os.waitpid(self.serverpid, 0)
    
    def getLinks(self, menu):
        links = []
        for i in xrange(len(menu)):
            if isinstance(menu[i], webvi.MenuItemLink):
                links.append(menu[i])
        return links
    
    def downloadMenuPage(self, reference, menuname):
        (status, statusmsg, menu) = self.client.get(reference)
        self.assertEqual(status, 200, 'Unexpected status code %s in %s menu' % (status, menuname))
        self.assertNotEqual(menu, None, 'Failed to get %s menu' % menuname)
        return menu

    def downloadAndExtractLinks(self, reference, minlinks, menuname):
        menu = self.downloadMenuPage(reference, menuname)
        links = self.getLinks(menu)
        self.assertTrue(len(links) >= minlinks, 'Too few links in %s menu' % menuname)
        return links

    def checkMediaUrl(self, reference):
        status, statusmsg, headers, body = self.client.get_raw(reference)
        self.assertEqual(status, 200, 'Unexpected status code %s while fetching mediaurl' % status)
        title, urls = self.client.parse_mediaurl(body)
        self.assertNotEqual(title, '', 'No title in mediaurl')
        self.assertTrue(len(urls) > 0, 'No URLs in mediaurl')

        # At least one of the urls should exists.
        OK = False
        for url in urls:
            res, reason = self.check_url(url)
            if res:
                OK = True
                break
        self.assertTrue(OK, 'No working URLs in mediaurl (last: %s)' % reason)

    def check_url(self, url):
        o = urlparse.urlsplit(url)
        if o.scheme != 'http':
            print 'Warning: Can\'t test scheme %s' % o.scheme
            return (True, '')
        
        p = ['']*5
        for i in range(2, 5):
            p[i] = o[i]
        path = urlparse.urlunsplit(p)

        conn = httplib.HTTPConnection(o.netloc, o.port)
        conn.request('HEAD', path)
        resp = conn.getresponse()
        conn.close()

        if resp.status in [301, 302, 303, 307]:
            redir = resp.getheader('Location', None)
            if redir is not None:
                return self.check_url(redir)

        return (True, '%d %s' % (resp.status, resp.reason))

    def moduleTest(self, refmain, minmainlinks=3, navigationlinkidx=2, minnavilinks=2, videolinkidx=0):
        # main category menu
        menu = self.client.get(refmain)
        self.assertNotEqual(menu, None, 'Failed to get category menu')

        links = self.getLinks(menu)
        self.assertTrue(len(links) >= minmainlinks, 'Too few links in category menu')

        # Navigation
        menu = self.client.get(links[navigationlinkidx].ref)
        self.assertNotEqual(menu, None, 'Failed to get navigation menu')
        
        links = self.getLinks(menu)
        self.assertTrue(len(links) >= minnavilinks, 'Too few links in navigation menu')

        # Video object
        self.assertNotEqual(links[videolinkidx].obj, None, 'No media object in a video link')

#         # Description page
#         videolink = links[videolinkidx].ref
#         self.assertNotEqual(videolink, None)
#         menu = self.client.get(videolink)
#         self.assertNotEqual(menu, None, 'Failed to get description page')        

    def testMainMenu(self):
        self.downloadAndExtractLinks('/mainmenu', 4, 'main')
        
    def testYoutube(self):
        # Category page
        links = self.downloadAndExtractLinks('youtube/categories.xsl,http://www.youtube.com/browse?s=mp', 3, 'category')

        # Navigation page
        # The third one is the first "proper" category. The first and second are "Search" and "All"
        navigationref = links[2].ref
        links = self.downloadAndExtractLinks(navigationref, 2, 'navigation')

        # Video link
        videolink = links[0]
        self.assertNotEqual(videolink.obj, None, 'No media object in a video link')
        self.checkMediaUrl(videolink.obj)
        

    def testYoutubeSearch(self):
        menu = self.downloadMenuPage('youtube/search.xsl,', 'search')
        self.assertTrue(len(menu) >= 5, 'Too few items in search menu')

        self.assertTrue(isinstance(menu[0], webvi.MenuItemTextField))
        self.assertTrue(isinstance(menu[1], webvi.MenuItemList))
        self.assertTrue(len(menu[1].items) >= 4)
        self.assertTrue(isinstance(menu[2], webvi.MenuItemList))
        self.assertTrue(len(menu[2].items) >= 4)
        self.assertTrue(isinstance(menu[3], webvi.MenuItemList))
        self.assertTrue(len(menu[3].items) >= 4)
        self.assertTrue(isinstance(menu[4], webvi.MenuItemSubmitButton))

        # Query term
        menu[0].value = 'youtube'
        # Sort by: rating
        menu[1].current = 3
        # Duration: Short
        menu[2].current = 1
        # Uploaded: This month
        menu[3].current = 3

        resultref = menu[4].activate()
        self.assertNotEqual(resultref, None)
        self.downloadAndExtractLinks(resultref, 1, 'search result')

    def testGoogleSearch(self):
        menu = self.downloadMenuPage('google/search.xsl,', 'search')
        self.assertTrue(len(menu) == 4, 'Unexpected number of items in Google search menu')

        self.assertTrue(isinstance(menu[0], webvi.MenuItemTextField))
        self.assertTrue(isinstance(menu[1], webvi.MenuItemList))
        self.assertTrue(len(menu[1].items) >= 4)
        self.assertTrue(isinstance(menu[2], webvi.MenuItemList))
        self.assertTrue(len(menu[2].items) >= 4)
        self.assertTrue(isinstance(menu[3], webvi.MenuItemSubmitButton))

        # Query term
        menu[0].value = 'google'
        # Sort by: date
        menu[1].current = 3
        # Duration: Short
        menu[2].current = 1

        resultref = menu[3].activate()
        self.assertNotEqual(resultref, None)
        self.downloadAndExtractLinks(resultref, 1, 'search result')

    def testSVTPlay(self):
        # Category page
        links = self.downloadAndExtractLinks('svtplay/categories.xsl,http://svtplay.se/kategorier', 3, 'category')

        # Navigation page
        navigationref = links[0].ref
        links = self.downloadAndExtractLinks(navigationref, 2, 'navigation')

        # Single program
        programref = links[0].ref
        links = self.downloadAndExtractLinks(programref, 1, 'program')

        # Video link
        videolink = links[0]
        self.assertNotEqual(videolink.obj, None, 'No media object in a video link')
        self.checkMediaUrl(videolink.obj)

    def testMetacafe(self):
        # Category page
        links = self.downloadAndExtractLinks('metacafe/categories.xsl,http://www.metacafe.com/videos/', 3, 'category')


        # Navigation page
        # The third one is the first "proper" category. The first and
        # second are "Search" and "Channels"
        navigationref = links[2].ref
        links = self.downloadAndExtractLinks(navigationref, 2, 'navigation')

        # Video link
        videolink = links[0]
        self.assertNotEqual(videolink.obj, None, 'No media object in a video link')
        self.checkMediaUrl(videolink.obj)

        # User channels
        links = self.downloadAndExtractLinks('metacafe/channellist.xsl,http://www.metacafe.com/api/channels/', 3, 'channel list')

    def testMetacafeSearch(self):
        menu = self.downloadMenuPage('metacafe/search.xsl,', 'search')
        self.assertTrue(len(menu) >= 4, 'Too few items in search menu')

        self.assertTrue(isinstance(menu[0], webvi.MenuItemTextField))
        self.assertTrue(isinstance(menu[1], webvi.MenuItemList))
        self.assertTrue(len(menu[1].items) == 3)
        self.assertTrue(isinstance(menu[2], webvi.MenuItemList))
        self.assertTrue(len(menu[2].items) == 4)
        self.assertTrue(isinstance(menu[3], webvi.MenuItemSubmitButton))

        # Query term
        menu[0].value = 'metacafe'
        # Sort by: most discussed
        menu[1].current = 2
        # Published: Anytime
        menu[2].current = 2

        resultref = menu[3].activate()
        self.assertNotEqual(resultref, None)
        self.downloadAndExtractLinks(resultref, 1, 'search result')


if __name__ == '__main__':
    # Run all tests
    unittest.main()

#     # Run single test
#     suite = unittest.TestSuite()
#     suite.addTest(TestServiceModules('testYoutube'))
#     unittest.TextTestRunner(verbosity=2).run(suite)
