/*
 * download.c: Web video plugin for the Video Disk Recorder
 *
 * See the README file for copyright information and how to reach the author.
 *
 * $Id$
 */

#include <unistd.h>
#include <sys/select.h>
#include <vdr/thread.h>
#include <vdr/tools.h>
#include "download.h"
#include "request.h"

// --- cCommunicationThread ------------------------------------------------
 
cCommunicationThread::cCommunicationThread() {
}

cCommunicationThread::~cCommunicationThread() {
  int numactive = activeList.Size();
  for (int i=0; i<activeList.Size(); i++) {
    delete activeList[i];
  }
  activeList.Clear();

  newListMutex.Lock();
  int numwaiting = newList.Size();
  for (int i=0; i<newList.Size(); i++) {
    delete newList[i];
  }
  newList.Clear();
  newListMutex.Unlock();

  finishedMutex.Lock();
  //int numfinished = finishedList.Size();
  for (int i=0; i<finishedList.Size(); i++) {
    delete finishedList[i];
  }
  finishedList.Clear();
  finishedMutex.Unlock();

  if ((numactive > 0) || (numwaiting > 0)) {
    esyslog("%d requests failed to complete, %d not even started", 
            numactive, numwaiting);
  }
}

void cCommunicationThread::Action(void) {
  fd_set readfds, writefds;
  int highfd, reqfd;
  struct timeval timeout;
  cDownloadRequest *req;

  while (Running()) {
    // Check if new requests should be starting
    newListMutex.Lock();
    for (int i=0; i<newList.Size(); i++) {
      req = newList[i];
      if (req->Aborted()) {
        // The request has been aborted even before we got a chance to
        // start it.
	MoveToFinishedList(req);
      } else {
        if (req->Connect() == -1) {
	  req->FinishRequest(-3, "Can't connect to daemon");
	  MoveToFinishedList(req);
	} else {
	  activeList.Append(req);
	}
      }
    }
    newList.Clear();
    newListMutex.Unlock();

    // Initialize fd sets
    FD_ZERO(&readfds);
    FD_ZERO(&writefds);
    highfd = 0;
    for (int j=0; j<activeList.Size(); j++) {
      req = activeList[j];
      int fd = req->GetFileNo();
      if (fd == -1) continue;
      if (fd > highfd)
        highfd = fd;
      if (req->Readable())
        FD_SET(fd, &readfds);
      if (req->Writable())
        FD_SET(fd, &writefds);
    }
    highfd = highfd+1;

    // Initialize the timeout data structure.
    timeout.tv_sec = 5;
    timeout.tv_usec = 0;

    int s = TEMP_FAILURE_RETRY(select(highfd, &readfds, &writefds, NULL, 
                                      &timeout));
    if (s == -1) {
      // select error
      LOG_ERROR_STR("select() error in downloader thread:");
    /*} else if (s == 0) {
      // timeout
      continue;*/
    }

    // Handle reads and writes on the available fds
    for (int j=0; j<activeList.Size(); j++) {
      req = activeList[j];

      if (req->Aborted()) {
        req->FinishRequest(-2, "Aborted");
      } else {
        reqfd = req->GetFileNo();

        if (FD_ISSET(reqfd, &readfds)) {
          req->Read();
        }
        if (FD_ISSET(reqfd, &writefds)) {
          req->Write();
        }
      }

      if (req->IsFinished()) {
        MoveToFinishedList(req);
      }
    }
  }
}

void cCommunicationThread::MoveToFinishedList(cDownloadRequest *req) {
  // Move the request from the activeList to finishedList.
  finishedMutex.Lock();
  for (int i=0; i<activeList.Size(); i++) {
    if (activeList[i] == req) {
      activeList.Remove(i);
      break;
    }
  }
  finishedList.Append(req);
  finishedMutex.Unlock();
}

void cCommunicationThread::AddRequest(cDownloadRequest *req) {
  newListMutex.Lock();
  newList.Append(req);
  newListMutex.Unlock();
}

void cCommunicationThread::Abort(long downloadID) {
  cFileDownloadRequest *req = (cFileDownloadRequest *)downloadID;

  // Check if the request is still waiting to be started.
  bool found = false;
  newListMutex.Lock();
  for (int i=0; i<newList.Size(); i++) {
    if (newList[i] == req) {
      req->Abort();

      found = true;
      break;
    }
  }
  newListMutex.Unlock();

  if (found)
    return;

  // Check if the request is among the active ones.
  finishedMutex.Lock();
  newListMutex.Lock();
  for (int i=0; i<activeList.Size(); i++) {
    if (activeList[i] == req) {
      req->Abort();
    }
  }
  newListMutex.Unlock();
  finishedMutex.Unlock();
}

cDownloadRequest *cCommunicationThread::GetFinishedRequest() {
  cDownloadRequest *res = NULL;
  finishedMutex.Lock();
  if (finishedList.Size() > 0) {
    res = finishedList[finishedList.Size()-1];
    finishedList.Remove(finishedList.Size()-1);
  }
  finishedMutex.Unlock();
  return res;
}

int cCommunicationThread::GetUnfinishedCount() {
  if (!Running())
    return 0;
  else
    return newList.Size() + activeList.Size();
}

int cCommunicationThread::GetUnfinishedFilesSummary(cVector<cFileDownloadSummary *> *summaries) {
  cFileDownloadRequest *filereq;
  int nfilereq = 0;

  // Items won't get added or deleted from activeList while we hold
  // both mutexes.
  finishedMutex.Lock();
  newListMutex.Lock();

  for (int i=0; i<activeList.Size(); i++) {
    filereq = dynamic_cast<cFileDownloadRequest *> (activeList[i]);
    if (filereq) {
      summaries->Append(filereq->CreateSummary((long)filereq));
      nfilereq++;
    }
  }

  newListMutex.Unlock();
  finishedMutex.Unlock();

  return nfilereq;
}
