/*
 * menu.c: Web video plugin for the Video Disk Recorder
 *
 * See the README file for copyright information and how to reach the author.
 *
 * $Id$
 */

#include <stdlib.h>
#include <time.h>
#include <vdr/skins.h>
#include <vdr/tools.h>
#include <vdr/i18n.h>
#include <vdr/osdbase.h>
#include <vdr/skins.h>
#include <vdr/font.h>
#include <vdr/osd.h>
#include "menu.h"
#include "download.h"
#include "config.h"
#include "common.h"
#include "parser.h"
#include "history.h"

// --- cSimpleLink ---------------------------------------------------------

cSimpleLink::cSimpleLink(const char *reference) {
  ref = reference ? strdup(reference) : NULL;
}

cSimpleLink::~cSimpleLink() {
  if (ref) {
    free(ref);
  }
}

char *cSimpleLink::GetURL() {
  return ref;
}

// --- cTextFieldData ------------------------------------------------------

cTextFieldData::cTextFieldData(int Length, const char *ID) {
  id = ID ? strdup(ID) : NULL;
  valuebufferlength = Length;
  valuebuffer = (char *)malloc(Length*sizeof(char));
  *valuebuffer = '\0';
}

cTextFieldData::~cTextFieldData() {
  if (id)
    free(id);
  if(valuebuffer)
    free(valuebuffer);
}

char *cTextFieldData::GetQueryFragment() {
  if (id && *id && valuebuffer) {
    char *encoded = URLencode(valuebuffer);
    cString tmp = cString::sprintf("%s=%s", id, encoded);
    free(encoded);
    return strdup(tmp);
  }

  return NULL;
}

char *cTextFieldData::GetValue() {
  return valuebuffer;
}

int cTextFieldData::GetLength() {
  return valuebufferlength;
}

// --- cItemListData -------------------------------------------------------

cItemListData::cItemListData(char **Strings, char **StringValues, int NumStrings, const char *ID) {
  strings = Strings;
  stringvalues = StringValues;
  numstrings = NumStrings;
  value = 0;
  id = ID ? strdup(ID) : NULL;
}

cItemListData::~cItemListData() {
  for (int i=0; i < numstrings; i++) {
    free(strings[i]);
    free(stringvalues[i]);
  }
  free(strings);
  free(stringvalues);
  if (id)
    free(id);
}

char *cItemListData::GetQueryFragment() {
  if (id && *id && (stringvalues[value][0] != '\0')) {
    cString tmp = cString::sprintf("%s=%s", id, stringvalues[value]);
    return strdup(tmp);
  }

  return NULL;
}

char **cItemListData::GetStrings() {
  return strings;
}

char **cItemListData::GetStringValues() {
  return stringvalues;
}

int cItemListData::GetNumStrings() {
  return numstrings;
}

int *cItemListData::GetValuePtr() {
  return &value;
}

// --- cSubmissionButtonData -----------------------------------------------

cSubmissionButtonData::cSubmissionButtonData(const char *queryUrl, 
					     cVector<cQueryData *> *objects) {
  querybase = queryUrl ? strdup(queryUrl) : NULL;
  queryObjects = objects;
}

cSubmissionButtonData::~cSubmissionButtonData() {
  if (querybase)
    free(querybase);
  // Do not delete queryObjects. It is owned by cNavigationMenu.
}

char *cSubmissionButtonData::GetURL() {
  if (!querybase)
    return NULL;

  char *querystr = (char *)malloc(sizeof(char)*(strlen(querybase)+2));
  strcpy(querystr, querybase);
  strcat(querystr, "?");

  int numparameters = 0;
  for (int i=0; i<queryObjects->Size(); i++) {
    cQueryData *qo = queryObjects->At(i);
    char *parameter = qo->GetQueryFragment();
    if (parameter) {
      querystr = (char *)realloc(querystr, (strlen(querystr)+strlen(parameter)+2)*sizeof(char));
      if (i > 0)
	strcat(querystr, "&");
      strcat(querystr, parameter);
      numparameters++;
    }
  }

  if (numparameters == 0) {
    // remove the '?' because no parameters were added to the url
    querystr[strlen(querystr)-1] = '\0';
  }

  return querystr;
}

// --- cXMLMenu --------------------------------------------------

cXMLMenu::cXMLMenu(const char *Title, int c0, int c1, int c2,
				       int c3, int c4)
: cOsdMenu(Title, c0, c1, c2, c3, c4)
{
}

void cXMLMenu::Deserialize(xmlDocPtr doc, xmlNodePtr node) {
  if (node)
    node = node->xmlChildrenNode;
  while (node) {
    if (node->type == XML_ELEMENT_NODE) {
      if (!CreateItemFromTag(doc, node)) {
        warning("Failed to parse menu tag: %s", (char *)node->name);
      }
    }
    node = node->next;
  }
}

int cXMLMenu::Load(const char *xmlstr) {
  Clear();

  xmlDocPtr xmldoc = xmlParseMemory(xmlstr, strlen(xmlstr));
  if (!xmldoc) {
    error("Failed to load menu.");

    xmlErrorPtr xmlerr =  xmlGetLastError();
    if (xmlerr) {
      error("libxml error: %s", xmlerr->message);
    }

    return -1;
  }

  Deserialize(xmldoc, xmlDocGetRootElement(xmldoc));

  xmlFreeDoc(xmldoc);
  return 0;
}


// --- cNavigationMenu -----------------------------------------------------

cNavigationMenu::cNavigationMenu(cNavigationMenu **parentMenuPointer, 
				 cHistory *History, cCommunicationThread *dlthread)
  : cXMLMenu("", 25)
{
  history = History;
  downloadthread = dlthread;
  pluginMenuPointer = parentMenuPointer;
  statusscreen = NULL;
  UpdateHelp();
}

cNavigationMenu::~cNavigationMenu() {
  *pluginMenuPointer = NULL;
  Clear();
}

bool cNavigationMenu::CreateItemFromTag(xmlDocPtr doc, xmlNodePtr node) {
  if (!xmlStrcmp(node->name, BAD_CAST "link")) {
    NewLinkItem(doc, node);
    return true;
  } else if (!xmlStrcmp(node->name, BAD_CAST "textfield")) {
    NewTextField(doc, node);
    return true;
  } else if (!xmlStrcmp(node->name, BAD_CAST "itemlist")) {
    NewItemList(doc, node);
    return true;
  } else if (!xmlStrcmp(node->name, BAD_CAST "textarea")) {
    NewTextArea(doc, node);
    return true;
  } else if (!xmlStrcmp(node->name, BAD_CAST "button")) {
    NewButton(doc, node);
    return true;
  } else if (!xmlStrcmp(node->name, BAD_CAST "title")) {
    NewTitle(doc, node);
    return true;
  }

  return false;
}

void cNavigationMenu::AddLinkItem(cOsdItem *item, cLinkBase *ref, 
                                  cLinkBase *objectref, cQueryData *queryObject) {
  Add(item);

  if (ref)
    links.Append(ref);
  else
    links.Append(NULL);

  if (objectref)
    objectLinks.Append(objectref);
  else
    objectLinks.Append(NULL);

  if (queryObject)
    queryObjects.Append(queryObject);
}

void cNavigationMenu::NewLinkItem(xmlDocPtr doc, xmlNodePtr node) {
  // label, ref and object tags
  xmlChar *itemtitle = NULL, *ref = NULL, *objectref = NULL;

  node = node->xmlChildrenNode;
  while (node) {
    if (!xmlStrcmp(node->name, BAD_CAST "label")) {
      if (itemtitle)
        xmlFree(itemtitle);
      itemtitle = xmlNodeListGetString(doc, node->xmlChildrenNode, 1);
    } else if (!xmlStrcmp(node->name, BAD_CAST "ref")) {
      if (ref)
        xmlFree(ref);
      ref = xmlNodeListGetString(doc, node->xmlChildrenNode, 1);
    } else if (!xmlStrcmp(node->name, BAD_CAST "object")) {
      if (objectref)
        xmlFree(objectref);
      objectref = xmlNodeListGetString(doc, node->xmlChildrenNode, 1);
    }
    node = node->next;
  }
  if (!itemtitle)
    itemtitle = xmlCharStrdup("???");
  debug("NewLinkItem: %s", itemtitle);

  const char *titleconv = csc.Convert((char *)itemtitle);
  cOsdItem *item = new cOsdItem(titleconv);
  cSimpleLink *objlinkdata = NULL;
  cSimpleLink *linkdata = NULL;
  if (ref)
    linkdata = new cSimpleLink((char *)ref);
  if (objectref) {
    // media object
    objlinkdata = new cSimpleLink((char *)objectref);
  } else {
    // navigation link
    char *bracketed = (char *)malloc((strlen(titleconv)+3)*sizeof(char));
    if (bracketed) {
      bracketed[0] = '\0';
      strcat(bracketed, "[");
      strcat(bracketed, titleconv);
      strcat(bracketed, "]");
      item->SetText(bracketed, false);
    }
  }
  AddLinkItem(item, linkdata, objlinkdata, NULL);

  xmlFree(itemtitle);
  if (ref)
    xmlFree(ref);
  if (objectref)
    xmlFree(objectref);
}

void cNavigationMenu::NewTextField(xmlDocPtr doc, xmlNodePtr node) {
  // id attribute
  xmlChar *id = xmlGetProp(node, BAD_CAST "id");

  // label tag
  xmlChar *text = NULL;
  node = node->xmlChildrenNode;
  while (node) {
    if (!xmlStrcmp(node->name, BAD_CAST "label")) {
      if (text)
        xmlFree(text);
      text = xmlNodeListGetString(doc, node->xmlChildrenNode, 1);
    }
    node = node->next;
  }
  if (!text)
    text = xmlCharStrdup("???");
  debug("NewTextField: %s", text);

  cTextFieldData *data = new cTextFieldData(256, (char *)id);
  cMenuEditStrItem *item = new cMenuEditStrItem(csc.Convert((char *)text), 
                                                data->GetValue(), 
                                                data->GetLength());
  AddLinkItem(item, NULL, NULL, data);

  free(text);
  if (id)
    xmlFree(id);
}

void cNavigationMenu::NewItemList(xmlDocPtr doc, xmlNodePtr node) {
  // id attribute
  xmlChar *id = xmlGetProp(node, BAD_CAST "id");
  debug("NewItemList");

  // label and item tags
  xmlChar *text = NULL;
  cVector<xmlChar *> items;
  cVector<xmlChar *> itemvalues;
  node = node->xmlChildrenNode;
  while (node) {
    if (!xmlStrcmp(node->name, BAD_CAST "label")) {
      if (text)
        xmlFree(text);
      text = xmlNodeListGetString(doc, node->xmlChildrenNode, 1);
    } else if (!xmlStrcmp(node->name, BAD_CAST "item")) {
      xmlChar *str = xmlNodeListGetString(doc, node->xmlChildrenNode, 1);
      if (!str)
        str = xmlCharStrdup("???");
      xmlChar *strvalue = xmlGetProp(node, BAD_CAST "value");
      if (!strvalue)
        strvalue = xmlCharStrdup("");

      items.Append(str);
      itemvalues.Append(strvalue);
    }
    node = node->next;
  }
  if (!text)
    text = xmlCharStrdup("???");

  char **itemtable = (char **)malloc(items.Size()*sizeof(char *));
  char **itemvaluetable = (char **)malloc(items.Size()*sizeof(char *));
  for (int i=0; i<items.Size(); i++) {
    itemtable[i] = strdup(csc.Convert((char *)items.At(i)));
    itemvaluetable[i] = strdup((char *)itemvalues.At(i));
  }

  cItemListData *data = new cItemListData(itemtable, itemvaluetable, 
                                          items.Size(), (char *)id);
  cMenuEditStraItem *item = new cMenuEditStraItem(csc.Convert((char *)text), 
                                                  data->GetValuePtr(), 
                                                  data->GetNumStrings(), 
                                                  data->GetStrings());
  AddLinkItem(item, NULL, NULL, data);

  for (int i=0; i<items.Size(); i++) {
    xmlFree(items.At(i));
    xmlFree(itemvalues.At(i));
  }
  xmlFree(text);
  if (id)
    xmlFree(id);
}

void cNavigationMenu::NewTextArea(xmlDocPtr doc, xmlNodePtr node) {
  // label tag
  xmlChar *itemtitle = NULL;
  debug("NewTextArea");

  node = node->xmlChildrenNode;
  while (node) {
    if (!xmlStrcmp(node->name, BAD_CAST "label")) {
      if (itemtitle)
        xmlFree(itemtitle);
      itemtitle = xmlNodeListGetString(doc, node->xmlChildrenNode, 1);
    }
    node = node->next;
  }
  if (!itemtitle)
    return;

  const cFont *font = cFont::GetFont(fontOsd);
  cTextWrapper tw(csc.Convert((char *)itemtitle), font, cOsd::OsdWidth());
  for (int i=0; i < tw.Lines(); i++) {
    AddLinkItem(new cOsdItem(tw.GetLine(i), osUnknown, false), NULL, NULL, NULL);
  }

  xmlFree(itemtitle);
}

void cNavigationMenu::NewButton(xmlDocPtr doc, xmlNodePtr node) {
  // label and submission tags
  xmlChar *itemtitle = NULL, *submission = NULL;

  node = node->xmlChildrenNode;
  while (node) {
    if (!xmlStrcmp(node->name, BAD_CAST "label")) {
      if (itemtitle)
        xmlFree(itemtitle);
      itemtitle = xmlNodeListGetString(doc, node->xmlChildrenNode, 1);
    } else if (!xmlStrcmp(node->name, BAD_CAST "submission")) {
      if (submission)
        xmlFree(submission);
      submission = xmlNodeListGetString(doc, node->xmlChildrenNode, 1);
    }
    node = node->next;
  }
  if (!itemtitle)
    itemtitle = xmlCharStrdup("???");
  debug("NewButton: %s", itemtitle);

  cSubmissionButtonData *data = new cSubmissionButtonData((char *)submission, &queryObjects);
  const char *titleconv = csc.Convert((char *)itemtitle); // do not free
  char *newtitle = (char *)malloc((strlen(titleconv)+3)*sizeof(char));
  if (newtitle) {
    newtitle[0] = '\0';
    strcat(newtitle, "[");
    strcat(newtitle, titleconv);
    strcat(newtitle, "]");

    cOsdItem *item = new cOsdItem(newtitle);
    AddLinkItem(item, data, NULL, NULL);
    free(newtitle);
  }

  xmlFree(itemtitle);
  if (submission)
    xmlFree(submission);
}

void cNavigationMenu::NewTitle(xmlDocPtr doc, xmlNodePtr node) {
  xmlChar *newtitle = xmlNodeListGetString(doc, node->xmlChildrenNode, 1);
  debug("NewTitle: %s", newtitle);
  if (newtitle) {
    SetTitle(csc.Convert((char *)newtitle));
    xmlFree(newtitle);
  }
}

eOSState cNavigationMenu::ProcessKey(eKeys Key)
{
  int old = Current();
  eOSState state = cXMLMenu::ProcessKey(Key);

  //debug("cNavigationMenu::ProcessKey: current = %d, size = %d", Current(), links.Size());
  if ((Current() < 0) || (Current() >= links.Size()))
    return state;

  if (state == osUnknown) {
    switch (Key) {
    case kInfo:
      // The alternative link is active only when object links are
      // present.
      if (objectLinks.At(Current()))
	state = Select(links.At(Current()), LT_REGULAR);
      break;
    case kOk:
      // Primary action: download media object or, if not a media
      // link, follow the navigation link.
      if (objectLinks.At(Current()))
	state = Select(objectLinks.At(Current()), LT_OBJECT);
      else
	state = Select(links.At(Current()), LT_REGULAR);
      break;
    case kBlue:
      // Secondary action: start streaming if a media object
      if (objectLinks.At(Current())) {
	state = Select(objectLinks.At(Current()), LT_STREAMOBJECT);
      }
      break;
    case kRed:
      state = HistoryBack();
      break;
    case kGreen:
      state = HistoryForward();
      break;
    case kYellow:
      SetStatusScreen(new cStatusScreen(this, downloadthread));
      state = AddSubMenu(statusscreen);
      break;
    default:
      break;
    }
  } else {
    // If the key press caused the selected item to change, we need to
    // update the help texts.
    //
    // In cMenuEditStrItem key == kOk with state == osContinue
    // indicates leaving the edit mode. We want to update the help
    // texts in this case also.
    if ((old != Current()) || 
	((Key == kOk) && (state == osContinue))) {
      UpdateHelp();
    }
  }

  //debug("cNavigationMenu::ProcessKey, state = %d", state);
  return state;
}

eOSState cNavigationMenu::Select(cLinkBase *link, eLinkType type)
{
  if (!link) {
    return osContinue;
  }
  char *ref = link->GetURL();
  if (!ref) {
    error("link->GetURL() == NULL in cNavigationMenu::Select");
    return osContinue;
  }

  if (type == LT_OBJECT) {
    downloadthread->AddRequest(new cFileDownloadRequest(webvideoConfig->GetServer(),
                                                        webvideoConfig->GetPort(),
                                                        history->Current()->GetID(),
                                                        ref, 
                                                        webvideoConfig->GetDownloadPath()));
    Skins.Message(mtInfo, tr("Downloading in the background"));
  } else if (type == LT_STREAMOBJECT) {
    downloadthread->AddRequest(new cStreamUrlRequest(webvideoConfig->GetServer(),
						     webvideoConfig->GetPort(),
						     history->Current()->GetID(),
						     ref));
    Skins.Message(mtInfo, tr("Starting player..."));
    return osEnd;
  } else {
    downloadthread->AddRequest(new cPageRequest(webvideoConfig->GetServer(),
                                                webvideoConfig->GetPort(),
                                                history->Current()->GetID(),
                                                ref));
    Skins.Message(mtStatus, tr("Retrieving..."));
  }

  return osContinue;
}

void cNavigationMenu::Clear(void) {
  cXMLMenu::Clear();
  SetTitle("");
  for (int i=0; i < queryObjects.Size(); i++) {
    delete queryObjects[i];
  }
  queryObjects.Clear();
  for (int i=0; i < links.Size(); i++) {
    if (links[i])
      delete links[i];
    if (objectLinks[i])
      delete objectLinks[i];
  }
  links.Clear();
  objectLinks.Clear();
}

void cNavigationMenu::Populate(cHistoryObject *page, const char *statusmsg) {
  Load(page->GetOSD());

  // Make sure that an item is selected (if there is at least
  // one). The help texts are not updated correctly if no item is
  // selected.
  SetCurrent(Get(0));
  UpdateHelp();
  SetStatus(statusmsg);
}

eOSState cNavigationMenu::HistoryBack() {
  cHistoryObject *page = history->Back();
  if (page) {
    Populate(page);
    Display();
  }
  return osContinue;
}

eOSState cNavigationMenu::HistoryForward() {
  cHistoryObject *before = history->Current();
  cHistoryObject *after = history->Forward();

  // Update only if the menu really changed
  if (before != after) {
    Populate(after);
    Display();
  }
  return osContinue;
}

void cNavigationMenu::UpdateHelp() {
  const char *red = (history->Current() != history->First()) ? tr("Back") : NULL;
  const char *green = (history->Current() != history->Last()) ? tr("Forward") : NULL;
  const char *yellow = tr("Status");
  const char *blue = ((Current() >= 0) && (objectLinks.At(Current()))) ? tr("Play") : NULL;

  SetHelp(red, green, yellow, blue);
}

cStatusScreen *cNavigationMenu::GetStatusScreen() {
  if (HasSubMenu()) {
    return statusscreen;
  } else {
    return NULL;
  }
}

// --- cStatusScreen -------------------------------------------------------

cStatusScreen::cStatusScreen(cNavigationMenu* parent, 
                             cCommunicationThread *thread)
  : cOsdMenu(tr("Unfinished downloads"), 40)
{
  parentmenu = NULL;
  downloadthread = thread;
  SetHelp(NULL, NULL, tr("Abort"), NULL);

  int charsperline = cOsd::OsdWidth() / cFont::GetFont(fontOsd)->Width('M');
  SetCols(charsperline-5);

  Update();
}

cStatusScreen::~cStatusScreen() {
  for (int i=0; i<summaries.Size(); i++) {
    delete summaries[i];
  }
  summaries.Clear();
}

void cStatusScreen::Update() {
  int c = Current();

  debug("Updating status screen");

  Clear();
  for (int i=0; i<summaries.Size(); i++) {
    delete summaries[i];
  }
  summaries.Clear();

  downloadthread->GetUnfinishedFilesSummary(&summaries);
  if (summaries.Size() == 0) {
    SetTitle(tr("No active downloads"));
  } else {
    for (int i=0; i<summaries.Size(); i++) {
      cFileDownloadSummary *s = summaries[i];
      cString dltitle = cString::sprintf("%s\t%3d%%", s->GetTitle(), 
					 (int)(100*s->GetPercentage()+0.5));
      Add(new cOsdItem(dltitle));
    }

    if (c >= 0)
      SetCurrent(Get(c));
  }

  lastupdate = time(NULL);

  Display();
}

bool cStatusScreen::NeedsUpdate() {
  return (Count() > 0) && (time(NULL) - lastupdate >= updateInterval);
}

eOSState cStatusScreen::ProcessKey(eKeys Key) {
  eOSState state = cOsdMenu::ProcessKey(Key);

  if (state == osUnknown) {
    switch (Key) {
    case kYellow:
      if ((Current() >= 0) && (Current() < summaries.Size())) {
        downloadthread->Abort(summaries[Current()]->GetDownloadID());
	delete summaries[Current()];
        summaries.Remove(Current());
        Del(Current());
        Display();
      }
      return osContinue;
    default:
      break;
    }
  }

  return state;
}
