/*
 * parser.c: Web video plugin for the Video Disk Recorder
 *
 * See the README file for copyright information and how to reach the author.
 *
 * $Id$
 */

#include <string.h>
#include <libxml/parser.h>
#include <vdr/tools.h>
#include "parser.h"
#include "common.h"

cCharSetConv csc = cCharSetConv("UTF-8", cCharSetConv::SystemCharacterTable());

// --- cUrlPriority --------------------------------------------------------

cUrlPriority::cUrlPriority(const char *URL, int Priority) {
  url = strdup(URL);
  priority = Priority;
}

cUrlPriority::~cUrlPriority() {
  free(url);
}

// --- cParser -------------------------------------------------------------

int cParser::ComparePriorities(const void *a, const void *b)
{
  return (*((cUrlPriority **)a))->GetPriority() - \
    (*((cUrlPriority **)b))->GetPriority();
}

eParsingError cParser::ParseMediaUrl(const char *page, int size, 
                                     cStringList *urls, char **title) {
  xmlChar *tmp = xmlCharStrndup(page, size);
  xmlDocPtr doc = xmlParseDoc(tmp);
  xmlFree(tmp);
  if (!doc) {
    xmlErrorPtr xmlerr = xmlGetLastError();
    error("Failed to parse xml document: %s", xmlerr->message);
    return PARSING_XML_ERROR;
  }

  if (title)
    *title = NULL;
  cVector<cUrlPriority *> priorities;
  
  xmlNodePtr node = xmlDocGetRootElement(doc);
  if (node) {
    if (xmlStrcmp(node->name, BAD_CAST "mediaurl")) {
      warning("root node is not <mediaurl>");
      xmlFreeDoc(doc);
      return PARSING_XML_ERROR;
    }

    node = node->xmlChildrenNode;
  }
  while(node) {
    if (!xmlStrcmp(node->name, BAD_CAST "title")) {
      if (title) {
	if (*title) {
	  free(*title);
	  *title = NULL;
	}
        xmlChar *nodetitle = xmlNodeListGetString(doc, node->xmlChildrenNode, 1);
	if (nodetitle) {
	  *title = strdup(csc.Convert((char *)nodetitle));
	  xmlFree(nodetitle);
	}
      }
    } else if (!xmlStrcmp(node->name, BAD_CAST "url")) {
      long prior = 50;
      xmlChar *priorstr = xmlGetProp(node, BAD_CAST "priority");
      if (priorstr) {
	prior = strtol((char *)priorstr, NULL, 10);
	xmlFree(priorstr);
      }

      xmlChar *url = xmlNodeListGetString(doc, node->xmlChildrenNode, 1);
      if (url) {
	priorities.Append(new cUrlPriority((char *)url, prior));
	xmlFree(url);
      } else {
	warning("Empty URL in mediaurl object.");
      }
    }
    node = node->next;
  }
  xmlFreeDoc(doc);

  if (title && !*title)
    *title = strdup("???");

  // Sort according to priorities (highest last)
  priorities.Sort(cParser::ComparePriorities);
  for (int i=0; i<priorities.Size(); i++) {
    urls->Append(strdup(priorities[i]->GetUrl()));
    delete priorities[i];
  }
  priorities.Clear();

  return PARSING_OK;
}
