/*
 * request.c: Web video plugin for the Video Disk Recorder
 *
 * See the README file for copyright information and how to reach the author.
 *
 * $Id$
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <netdb.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>
#include <time.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <vdr/tools.h>
#include <vdr/i18n.h>
#include "request.h"
#include "buffer.h"
#include "mimetypes.h"
#include "common.h"
#include "parser.h"
#include "config.h"

// --- cFileDownloadSummary ------------------------------------------------

cFileDownloadSummary::cFileDownloadSummary(const char *title, 
                                           int downloadedbytes, 
                                           int totalbytes,
                                           long downloadid) {
  name = title ? strdup(title) : NULL;
  downloaded = downloadedbytes;
  total = totalbytes;
  id = downloadid;
}

cFileDownloadSummary::~cFileDownloadSummary() {
  if (name)
    free(name);
}

const char *cFileDownloadSummary::GetTitle() {
  if (name)
    return name;
  else
    return "???";
}

float cFileDownloadSummary::GetPercentage() {
  if ((downloaded < 0) || (total < 0))
    return 0.0;
  else
    return float(downloaded)/total;
}

// --- cDownloadRequest ----------------------------------------------------

cDownloadRequest::cDownloadRequest(const char *serveraddr, 
                                   const char *serverport, int ID) {
  server = serveraddr ? strdup(serveraddr) : NULL;
  port = serverport ? strdup(serverport) : NULL;
  termstr = NULL;
  termbytes = 0;
  termIsString = false;
  contentLength = -1;
  contentType = NULL;
  lastbody = NULL;
  state = STATUSLINE;
  statusCode = -1;
  statusPharse = NULL;
  reqID = ID;
  fd = -1;
  aborted = false;
}

cDownloadRequest::~cDownloadRequest() {
  Close();
  outBuffers.Clear();
  if (termstr)
    free(termstr);
  if (lastbody)
    free(lastbody);
  if (contentType)
    free(contentType);
  if (statusPharse)
    free(statusPharse);
  if (server)
    free(server);
  if (port)
    free(port);
}

int cDownloadRequest::TryConnect() {
  struct addrinfo hints, *servinfo, *si;
  int status;
  int sfd, connectedfd = -1;

  memset(&hints, 0, sizeof(hints));
  hints.ai_family = AF_UNSPEC;
  hints.ai_socktype = SOCK_STREAM;

  if ((status = getaddrinfo(server, port, &hints, &servinfo)) != 0) {
    esyslog("getaddrinfo failed: %s", gai_strerror(status));
    return -1;
  }

  for (si = servinfo; si; si = si->ai_next) {
    sfd = socket(si->ai_family, si->ai_socktype, si->ai_protocol);
    if (sfd == -1)
      continue;

    if (connect(sfd, si->ai_addr, si->ai_addrlen) != -1) {
      connectedfd = sfd;
      break;
    }
    
    close(sfd);
  }
  freeaddrinfo(servinfo);

  if (connectedfd == -1) {
    return -1;
  }

  // Set the socket in non-blocking mode
  fcntl(connectedfd, F_SETFL, fcntl(fd, F_GETFL) | O_NONBLOCK);

  return connectedfd;
}

bool cDownloadRequest::StartDaemon() {
  pid_t pid = fork();
  if (pid == 0) {
    // child
    
    // Close all file descriptors. VDR doesn't like it if the daemon
    // process keeps /dev/dvb/adapter0/dvr0 open.
    for (int fd = 3; fd < 512; fd++) {
      close(fd);
    }

    char * const *daemoncmd = webvideoConfig->GetDaemonCmd();
    if (!daemoncmd[0]) {
      error("no daemon command specified");
      _exit(1);
    }

    debug("executing daemon %s", daemoncmd[0]);
    for (int i=1; daemoncmd[i]; i++) {
      debug("%s", daemoncmd[i]);
    }
    int err = execv(daemoncmd[0], daemoncmd);
    if (err == -1) {
      LOG_ERROR_STR("Failed to execute daemon");
      _exit(1);
    }
    _exit(0);
  } else if (pid < 0) {
    // error
    LOG_ERROR_STR("Failed to fork");
    return false;
  } else {
    // parent
    int status;
    waitpid(pid, &status, 0); // don't leave zombies

    // wait a bit more for the daemon to start
    struct timespec tv;
    tv.tv_sec = 0;
    tv.tv_nsec = 500*1000*1000;  // 0.5 s
    nanosleep(&tv, NULL);
  }

  return true;
}

int cDownloadRequest::Connect() {
  // close the current connection
  Close();
  aborted = false;

  fd = TryConnect();
  if (fd == -1) {
    if (errno == ECONNREFUSED) {
      debug("Connection refused");
      // Perhaps the server is not running? Try to start it and
      // reconnect, but only if trying to connect to localhost.
      if ((!strcmp(server, "localhost")) || (!strcmp(server, "127.0.0.1"))) {
	isyslog("Daemon did not respond. Trying to start webvid.");
        if (StartDaemon()) {
	  debug("trying again");
          fd = TryConnect();
        }
      }
    }
  }
  debug("tryconnect fd = %d", fd);
  if (fd == -1) {
    LOG_ERROR;
  }

  return fd;
}

void cDownloadRequest::Close() {
  if (fd != -1) {
    outBuffers.Clear();
    SendRequest("CLOSE X WVTP/1.0\r\n\r\n", false);
    while (outBuffers.Size() > 0)
      Write();

    int s = TEMP_FAILURE_RETRY(close(fd));
    if (s == -1)
      LOG_ERROR;
    fd = -1;
  }
}

bool cDownloadRequest::Connected() {
  return fd != -1;
}

bool cDownloadRequest::Readable() {
  return Connected();
}

bool cDownloadRequest::Writable() {
  return Connected() && (outBuffers.Size() > 0);
}

void cDownloadRequest::Push(const char *send) {
  if (send)
    outBuffers.Append(strdup(send));
  else
    outBuffers.Append(NULL);
}

bool cDownloadRequest::AcceptPartialBody() {
  return false;
}

void cDownloadRequest::SetTerminator(const char *term) {
  termIsString = true;
  if (termstr)
    free(termstr);
  termstr = NULL;
  if (term)
    termstr = strdup(term);
}

void cDownloadRequest::SetTerminator(int bytes) {
  termIsString = false;
  termbytes = bytes;
}

bool cDownloadRequest::SearchTerminator() {
  // Searches for terminator condition in the input buffer and calls
  // TerminatorFound() if found. Returns true, if there may be more
  // terminators in the buffer still. In that case, the caller might
  // want to call us again before fetching more data from the network.
  if (termIsString) {
    int termlen = 0;
    char *found = NULL;
    if (!termstr || (*termstr == '\0')) {
      // If the terminator is empty always call TerminatorFound() for
      // the whole buffer
      termlen = 0;
      found = inBuffer.Get() + inBuffer.Length();
    } else {
      // Find the terminator string
      termlen = strlen(termstr);
      found = (char *)memmem(inBuffer.Get(), inBuffer.Length(), 
                             termstr, termlen);
    }
    if (found) {
      char *buf = inBuffer.Get();
      int len = found-buf;

      inBuffer.Pop(TerminatorFound(buf, len+termlen));
      return inBuffer.Length() >= termlen;
    }
  } else {
    if (termbytes <= 0) {
      // If term is zero, call TerminatorFound on the data as it comes
      // along.
      inBuffer.Pop(TerminatorFound(inBuffer.Get(), inBuffer.Length()));
      return inBuffer.Length() > 0;
    } else if (inBuffer.Length() >= termbytes) {
      // Numerical terminator means that we wait until at least
      // termbytes bytes has been received.
      inBuffer.Pop(TerminatorFound(inBuffer.Get(), termbytes));
      return inBuffer.Length() >= termbytes;
    }
  }

  return false;
}

// vdr already handles (ignores) SIGPIPE. We probably should not
// changes the signal handler here.

void cDownloadRequest::Read() {
  ssize_t s = inBuffer.PutFromFile(fd, inbufsize);

  if (s == -1) {
    // error
    int errsv = errno;
    LOG_ERROR;
    FinishRequest(-1, strerror(errsv));
    return;
  } else if (s == 0) {
    // (unexpected) end of stream
    fd = -1;
    if (AcceptPartialBody())
      FinishRequest(200, "OK, but message is shorter than Content-Length");
    else
      FinishRequest(-1, "Connection closed while trying to read");
    return;
  }

  // Handle all terminators currently in the buffer
  while (SearchTerminator()) {}
}

void cDownloadRequest::Write() {
  if (outBuffers.Size() == 0)
    return;

  char *out = outBuffers[0];
  outBuffers.Remove(0);

  if (out) {
    int s = TEMP_FAILURE_RETRY(write(fd, out, strlen(out)));
    if (s == -1) {
      // error
      int errsv = errno;
      LOG_ERROR;
      FinishRequest(-1, strerror(errsv));
      free(out);
    } else if (s < (int)strlen(out)) {
      // Only a substring was written. Put the buffer back to outBuffers.
      memmove(out, &out[s], strlen(out)-s+1);
      outBuffers.Insert(out, 0);
    } else {
      // writing succeeded
      free(out);
    }
  } else { // out == NULL
    // unexpected closing of connection
    fd = -1;
    FinishRequest(-1, "Connection closed while trying to write");
  }
}

int cDownloadRequest::HandleBody(const char *body, int bodylen) {
  if (lastbody)
    free(lastbody);
  lastbody = strndup(body, bodylen);

  if ((long)bodylen != contentLength) {
    warning("The length of the body (%d) not equal to Content-Length (%ld)", 
            bodylen, contentLength);
  }

  FinishRequest();

  return bodylen;
}

bool cDownloadRequest::Success() {
  return (statusCode == 0) || (statusCode == 200);
}

void cDownloadRequest::ParseStatusLine(const char *line, int len) {
  // The status line is:
  // WVTP/1.0 Status-Code Reason-Phrase

  statusCode = -1;
  if (statusPharse)
    free(statusPharse);
  statusPharse = NULL;

  if (strncmp(line, "WVTP/1.0 ", 9)) {
    // Unknown protocol or protocol version
    return;
  }

  // Skip the protocol version
  line = line+9;
  len = len-9;

  // Discard the line feed at the end
  len = len-2;  

  char *sep = (char *)memchr(line, ' ', len);
  if (sep) {
    char *endptr = NULL;
    int codelen = sep-line;
    char *tmp = strndup(line, codelen);
    statusCode = (int)strtol(tmp, &endptr, 10);

    if (*endptr != '\0') {
      // The string was not a number, fail.
      statusCode = -1;
      free(tmp);
      return;
    }
    free(tmp);

    statusPharse = strndup(&sep[1], len-codelen-1);
  }
}

int cDownloadRequest::TerminatorFound(const char *ptr, int len) {
  if (state == STATUSLINE) {
    // The first line of the response contains the status code
    ParseStatusLine(ptr, len);
    if (statusCode == -1)
      FinishRequest(-1, "Failed to parse status line");
    state = HEADERS;
  } else if (state == HEADERS) {
    if (len == 2) {
      // Headers end with an empty line
      if (contentLength != -1)
        SetTerminator(contentLength);
      else
        SetTerminator((const char *)NULL);
      state = BODY;
      HandleHeadersOver();
    } else {
      // A header line
      if (!strncasecmp(ptr, "Content-Length:", 15)) {
        char *tmp = strndup(&ptr[15], len-15-2);
        contentLength = strtol(tmp, NULL, 10);
        free(tmp);
      } else if (!strncasecmp(ptr, "Content-Type:", 13)) {
        if (contentType)
          free(contentType);
        contentType = compactspace(strndup(&ptr[13], len-13-2));
      }
      HandleHeader(ptr, len);
    }
  } else {
    return HandleBody(ptr, len);
  }

  return len;
}

void cDownloadRequest::HandleHeader(const char *header, int headerlen) {
  // Do nothing
}

void cDownloadRequest::HandleHeadersOver() {
  // On error, we can finish here because there will be no body.
  if (!Success()) {
    FinishRequest();
  }
}

void cDownloadRequest::SendRequest(const char *reqstr, bool resetStatus) {
  debug("SendRequest: %s", reqstr);

  if (resetStatus) {
    statusCode = -1;
    if (statusPharse) {
      free(statusPharse);
      statusPharse = NULL;
    }
  }

  contentLength = -1;
  state = STATUSLINE;
  SetTerminator("\r\n");
  Push(reqstr);
}

void cDownloadRequest::FinishRequest(int status, const char *errmsg) {
  debug("FinishRequest, status = %d, %s", status, errmsg);

  Close();
  state = FINISHED;

  if (status != 0) {
    statusCode = status;
  }
  if (errmsg) {
    if (statusPharse)
      free(statusPharse);
    statusPharse = strdup(errmsg);
  }

  if (!Success()) {
    error("Request %d failed: %d %s", reqID, status, errmsg);
  }
}

// --- cPageRequest --------------------------------------------------------

cPageRequest::cPageRequest(const char *serveraddr, const char *serverport, int ID, const char *ref)
  : cDownloadRequest(serveraddr, serverport, ID)
{
  SendRequest(cString::sprintf("GET %s WVTP/1.0\r\n\r\n", ref));
}

// --- cFileDownloadRequest ------------------------------------------------

cFileDownloadRequest::cFileDownloadRequest(const char *serveraddr, 
                                           const char *serverport,
					   int ID,
                                           const char *objecturl, 
                                           const char *destdirectory)
: cDownloadRequest(serveraddr, serverport, ID)
{
  destdir = strdup(destdirectory);
  destbasename = NULL;
  destfile = NULL;
  tempfilename = NULL;
  downloadMode = false;
  sourceUrl = NULL;
  bytesDownloaded = -1;
  SendRequest(cString::sprintf("GET %s WVTP/1.0\r\n\r\n", objecturl));
}

cFileDownloadRequest::~cFileDownloadRequest() {
  free(destdir);
  if (destbasename)
    free(destbasename);
  if (destfile)
    delete destfile;
  if (tempfilename)
    free(tempfilename);
  if (sourceUrl)
    free(sourceUrl);
}

bool cFileDownloadRequest::AcceptPartialBody() {
  // When downloading a file, too short message body is not considered
  // a serious error. Hopefully a media player can still play the
  // beginning of the file.
  debug("The body is less than Content-Length bytes, downloadmode = %d", downloadMode);
  return downloadMode;
}

void cFileDownloadRequest::DownloadNextUrl() {
  if (urls.Size() > 0) {
    if (sourceUrl)
      free(sourceUrl);
    sourceUrl = urls[urls.Size()-1];

    char *tmp = (char *)malloc((strlen(sourceUrl)+32)*sizeof(char));
    *tmp = '\0';
    strcat(tmp, "DOWNLOAD ");
    strcat(tmp, sourceUrl);
    strcat(tmp, " WVTP/1.0\r\n\r\n");
    SendRequest(tmp);
    free(tmp);

    urls.Remove(urls.Size()-1);
    downloadMode = true;
    bytesDownloaded = 0;
  } else {
    FinishRequest(404, "Failed to download any of the URLs");
  }
}

char *cFileDownloadRequest::GetExtension() {
  // Get extension from Content-Type
  char *ext = NULL;
  char *ext2 = MimeTypes->ExtensionFromMimeType(contentType);

  // Workaround for buggy servers: If the server claims that the mime
  // type is text/plain, ignore the server and fall back to extracting
  // the extension from the URL. This function should be called only
  // for video, audio or ASX files and therefore text/plain is clearly
  // incorrect.
  if (ext2 && contentType && !strcasecmp(contentType, "text/plain")) {
    debug("Ignoring content type text/plain, getting extension from url.");
    free(ext2);
    ext2 = NULL;
  }

  if (ext2) {
    // Append dot in the start of the extension
    ext = (char *)malloc(strlen(ext2)+2);
    ext[0] = '.';
    ext[1] = '\0';
    strcat(ext, ext2);
    free(ext2);
    return ext;
  }

  // FIXME

  // Get extension from URL
  ext = extensionFromUrl(sourceUrl);
  if (ext) {
    return ext;
  }

  // No extension!
  return strdup("");
}

const char *cFileDownloadRequest::GetTempFileName() {
  return tempfilename;
}

char *cFileDownloadRequest::GetFinalFileName() {
  char *ext = GetExtension();
  cString dest = cString::sprintf("%s/%s%s", destdir, destbasename, ext);

  struct stat st;
  int i = 1;
  while (stat(dest, &st) == 0) {
    dest = cString::sprintf("%s/%s-%d%s", destdir, destbasename, i++, ext);
  }
  free(ext);

  if (errno == ENOENT) {
    // dest is an available name
    return strdup(dest);
  } else {
    // Error accessing the dest dir (for example, no search
    // permissions or too many symbolic links)
    return NULL;
  }
}

void cFileDownloadRequest::FinishRequest(int err, const char *errmsg) {
  cDownloadRequest::FinishRequest(err, errmsg);

  if (destfile) {
    delete destfile;
    destfile = NULL;

    if (Success()) {
      // Copy the temporary file to the final destination.
      const char *tempname = GetTempFileName();
      char *finalname = GetFinalFileName();
      debug("moving %s to %s", tempname, finalname);
      if ((!finalname) || (moveFile(tempname, finalname) != 0))
        LOG_ERROR_STR("Failed to rename the temporary file");
      if (finalname)
        free(finalname);
    } else {
      // Remove the temporary file
      debug("Removing temporary file %s", GetTempFileName());
      unlink(GetTempFileName());
    }
  }
}

void cFileDownloadRequest::HandleHeadersOver() {
  if (downloadMode) {
    if (!Success()) {
      // DOWNLOAD failed. Try the next URL.
      debug("Download failed with error %d", statusCode);
      DownloadNextUrl();
    } else if (contentLength == -1) {
      // There must be a Content-Length header.
      DownloadNextUrl();
    } else {
      // Headers fine. Write the body to a temporary file.
      if (tempfilename)
        free(tempfilename);
      tempfilename = tempnam(destdir, "webvi");
      destfile = new cUnbufferedFile();
      if (destfile->Open(tempfilename, O_WRONLY | O_CREAT | O_EXCL, DEFFILEMODE) < 0) {
        error("Failed to create temporary file %s", tempfilename);
        delete destfile;
        destfile = NULL;
	FinishRequest(-1, "Failed to create a temporary file");
      }

      // Do not buffer the body, but write it to the destfile
      // immediately.
      SetTerminator(0);
    }
  } else {
    cDownloadRequest::HandleHeadersOver();
  }
}

int cFileDownloadRequest::HandleBody(const char *body, int bodylen) {
  if (downloadMode) {
    // download mode: this is the file content.
    if (destfile) {
      ssize_t s = destfile->Write(body, bodylen);
      bytesDownloaded += s;
      if (s == -1) {
        // error writing to file
        int errsv = errno;
        LOG_ERROR;
        FinishRequest(-1, strerror(errsv));
        return bodylen;
      }
      if (bytesDownloaded >= contentLength) {
	// The whole file has been downloaded
	FinishRequest();
      }
      return s;
    }
  } else {
    // This is the reply to the initial GET. Parse the results, and send
    // the DOWNLOAD request.
    urls.Clear();
    cParser::ParseMediaUrl(body, bodylen, &urls, &destbasename);
    destbasename = safeFilename(destbasename);
    DownloadNextUrl();
  }

  return bodylen;
}

cFileDownloadSummary *cFileDownloadRequest::CreateSummary(long downloadid) {
  const char *filename;
  if (destbasename)
    filename = destbasename;
  else
    filename = "???";

  cString title;
  if (Aborted()) {
    title = cString::sprintf("[%s] %s", tr("Aborted"), filename);
  } else {
    title = cString(filename);
  }

  return new cFileDownloadSummary(title, bytesDownloaded, contentLength, 
                                  downloadid);
}


// --- cStreamUrlRequest ---------------------------------------------------

cStreamUrlRequest::cStreamUrlRequest(const char *serveraddr, 
                                     const char *serverport, 
                                     int ID, const char *ref) 
  : cDownloadRequest(serveraddr, serverport, ID)
{
  GETmode = true;
  SendRequest(cString::sprintf("GET %s WVTP/1.0\r\n\r\n", ref));
}

void cStreamUrlRequest::StreamNextUrl() {
  char *currentUrl;

  if (streamUrls.Size() > 0) {
    GETmode = false;
    currentUrl = streamUrls[streamUrls.Size()-1];
    char *tmp = (char *)malloc((strlen(currentUrl)+32)*sizeof(char));
    *tmp = '\0';
    strcat(tmp, "STREAM ");
    strcat(tmp, currentUrl);
    strcat(tmp, " WVTP/1.0\r\n\r\n");
    SendRequest(tmp);
    free(tmp);
    streamUrls.Remove(streamUrls.Size()-1);
  } else {
    FinishRequest(404, "No URLs in the GET response");
  }
}

void cStreamUrlRequest::HandleHeadersOver() {
  if (GETmode) {
    cDownloadRequest::HandleHeadersOver();
  } else {  // STREAM mode
    if (statusCode == 404)
      StreamNextUrl();
    else
      cDownloadRequest::HandleHeadersOver();
  }
}

int cStreamUrlRequest::HandleBody(const char *body, int bodylen) {
  if (GETmode) {
    // Reply to initial GET
    streamUrls.Clear();
    cParser::ParseMediaUrl(body, bodylen, &streamUrls, NULL);
    StreamNextUrl();
    return bodylen;
  } else {
    // Reply to STREAM
    return cDownloadRequest::HandleBody(body, bodylen);
  }
}
