/*
 * request.h: Web video plugin for the Video Disk Recorder
 *
 * See the README file for copyright information and how to reach the author.
 *
 * $Id$
 */

#ifndef __WEBVIDEO_REQUEST_H
#define __WEBVIDEO_REQUEST_H

#include <vdr/tools.h>
#include "buffer.h"

enum eRequestType { REQT_NONE, REQT_MENU, REQT_FILE, REQT_STREAM };
enum eRequestState { STATUSLINE, HEADERS, BODY, FINISHED };

// --- cFileDownloadSummary ------------------------------------------------

class cFileDownloadSummary {
private:
  char *name;
  int downloaded;
  int total;
  long id;
public:
  cFileDownloadSummary(const char *title, int downloadedbytes, 
                       int totalbytes, long downloadid);
  ~cFileDownloadSummary();

  long GetDownloadID() { return id; };
  const char *GetTitle();
  float GetPercentage();
};

// --- cDownloadRequest ----------------------------------------------------

class cDownloadRequest {
private:
  int reqID;
  char *server;
  char *port;
  int fd;

  bool aborted;

  char *termstr;
  int termbytes;
  bool termIsString;

  static const int inbufsize = 4096;
  cMemoryBuffer inBuffer;
  cStringList outBuffers;

  char *lastbody;

  int TryConnect();
  bool SearchTerminator();
protected:
  eRequestState state;
  long contentLength;
  char *contentType;
  int statusCode;
  char *statusPharse;

  void ParseStatusLine(const char *line, int len);
  virtual bool AcceptPartialBody();
public:
  cDownloadRequest(const char *serveraddr, const char *port, int ID);
  virtual ~cDownloadRequest();

  int GetID() { return reqID; };
  int GetFileNo() { return fd; };

  int Connect();
  void Close();
  bool Connected();
  bool StartDaemon();
  virtual bool Readable();
  virtual bool Writable();

  // Signal that the request has been finished. Reset the status code
  // to err and the status pharse to errmsg. The default values err ==
  // 0 and errmsg == NULL keep the current status code and pharse
  // unchanged.
  virtual void FinishRequest(int err=0, const char *errmsg=NULL);
  bool IsFinished() { return state == FINISHED; };
  void Abort() { aborted = true; };
  bool Aborted() { return aborted; };

  // Store copy of send into output FIFO. If send is NULL, close the
  // connection when all items previously in the output FIFO has been
  // sent.
  void Push(const char *send);
  // Push request to output FIFO, and reset the internal state
  void SendRequest(const char *request, bool resetStatus = true);

  // The terminator is searched in the incoming data. When a
  // terminator is found, TerminatorFound() is called for all data
  // read so far. Terminator can be either a string, or a byte count.
  void SetTerminator(const char *termstr);
  void SetTerminator(int termbytes);
  // Called when a terminator is found on the input buffer. ptr points
  // to the start of the buffer, and len is the size of the datablock
  // in the buffer including the terminator. Returns the number of
  // bytes which were consumed.
  virtual int TerminatorFound(const char *ptr, int len);

  // Call when there is incoming data in the file descriptor GetFileNo().
  virtual void Read();
  // Call when the file descriptor GetFileNo() is ready for writing.
  virtual void Write();
  
  // Called for each header line.
  virtual void HandleHeader(const char *header, int headerlen);
  // Called after all headers has been handled.
  virtual void HandleHeadersOver();
  // Called when a terminator is found and we are not receiving
  // headers. Return the number of bytes consumed.
  virtual int HandleBody(const char *body, int bodylen);

  // Return true if the lastest status code indicates success.
  bool Success();
  // Return the status code of the current request.
  int GetStatusCode() { return statusCode; };
  // Return the response pharse of the current request.
  const char *GetStatusPharse() { return statusPharse; };
  // Return true if the response body is a XML-representation of a
  // menu.
  virtual eRequestType GetType() { return REQT_NONE; };
  // Return the body of the last succeeded request.
  virtual const char *GetLastBody() { return lastbody; };
};

// --- cPageRequest --------------------------------------------------------

class cPageRequest : public cDownloadRequest {
private:
public:
  cPageRequest(const char *serveraddr, const char *port, int ID, const char *ref);
  virtual eRequestType GetType() { return REQT_MENU; };
};

// --- cFileDownloadRequest ------------------------------------------------

class cFileDownloadRequest : public cDownloadRequest {
private:
  char *sourceUrl;
  long bytesDownloaded;
  char *destdir;
  char *destbasename;
  char *tempfilename;
  cUnbufferedFile *destfile;
  cStringList urls;
  bool downloadMode;
protected:
  char *GetExtension();
  const char *GetTempFileName();
  char *GetFinalFileName();
  void DownloadNextUrl();
  bool AcceptPartialBody();
public:
  cFileDownloadRequest(const char *serveraddr, const char *port, 
                       int ID, const char *objecturl, 
                       const char *destdir);
  ~cFileDownloadRequest();

  void FinishRequest(int err=0, const char *errmsg=NULL);
  void HandleHeadersOver();
  int HandleBody(const char *body, int bodylen);

  virtual eRequestType GetType() { return REQT_FILE; };

  cFileDownloadSummary *CreateSummary(long downloadid);
};

// --- cStreamUrlRequest ---------------------------------------------------

class cStreamUrlRequest : public cDownloadRequest {
private:
  bool GETmode;
  cStringList streamUrls;

  void StreamNextUrl();
public:
  cStreamUrlRequest(const char *serveraddr, const char *port, 
                    int ID, const char *playlisturl);

  void HandleHeadersOver();
  int HandleBody(const char *body, int bodylen);

  virtual eRequestType GetType() { return REQT_STREAM; };
};

#endif // __WEBVIDEO_REQUEST_H
